static char RcsId[] = "$Header: /data1/src/duptape/RCS/recordio.c,v 1.13 1994/03/04 19:26:41 will Exp $" ;

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "recordio-vers.h"

#ifdef __STDC__
#include <stdlib.h>
#else
#include <malloc.h>
#endif

#define PROTOCOL_LEADER	"RCIO"	/* don't change this */
#define PROTOCOL_VERS	"0002"	/* exactly 4 chars; increment iff protocol
				 * is changed.
				 */

/* recordio: record-length-preserving I/O.
 *	Useful for duplicating tapes:
 *	recordio -r	reads a tape from stdin and writes to stdout:
 *			"RCIO"
 *			<vers>	(4 chars)
 *			<bufsize> (8 char str)	<- --------
 *			<reclen1> <record1>	<- This is the vers="0002" fmt.
 *			<reclen2> <record2>	<- --------
 *			....
 *	recordio -w	reads from stdin; contents must be in recordio -r fmt.
 *			It mallocs a buffer <bufsize> long, and then copies
 *			the record data to stdout, preserving record lengths.
 * recordio -r options:
 *		-b bufsize copy buffer size.
 *		-n nfm	number of files to copy.  Default is to double EOF.
 *			After each file but the last is copied, an EOF is
 *			written.  (recordio -r indicates an EOF by inserting
 *			a recordlength of 0.)
 *		-z	Don't copy zero-length files (ie multiple EOF's).
 * recordio -r|-w options:
 *		-v verbose about #EOF's read or written.
 *		-V very verbose: info about records as well as EOF's.

 *	No EOF is explicitly added after the last file, on the assumption
 *	that the device driver will write a terminating EOF upon close.

 */

#include <sys/types.h>

/* For putting EOF's */
#ifdef _AIX
#include <sys/tape.h>
#else
#include <sys/mtio.h>
#endif
#include <sys/ioctl.h>

#define DEFAULT_BUFSIZE 32768

char *optarg;
int optind, opterr;

char *prog;

#ifdef __STDC__
    void Error(int, int, char *, ...);
    void usage(void);
#else
    void Error();
    void usage();
#endif


void
usage()
{
    (void) fprintf(stderr,
	    "Use: %s -r [-b bufsize] [-n nfm] [-z] [-v] [-V] < tapedevice \n\
\t ... | %s -w [-v] > anothertapedevice\n",
		prog, prog);
}

int
main(argc, argv)
int argc;
char **argv;
{
    void readin(), writeout();
    int bufsize=-1;
    int readtape=0;
    int writetape=0;
    int copyzerolength=1;
    int nfm=-1;
    int i;
    int verbose=0;
    char *s;

    prog = ((s = strrchr(argv[0], '/')) && *(s+1))  ? s+1 : argv[0];

    while ((i = getopt(argc, argv, "hrvVwzn:b:")) != -1) {
	switch (i) {
	case 'b':
	    bufsize = atoi(optarg);
	    if (bufsize <= 0)
		Error(0, 1, "%s: negative buffer sizes (%d) not allowed",
			prog, bufsize);
	    break;
	case 'h':
	    usage();
	    (void) exit(0);
	    break;
	case 'r':
	    readtape = 1;
	    break;
	case 'v':
	    verbose=1;
	    break;
	case 'V':
	    verbose=2;
	    break;
	case 'w':
	    writetape = 1;
	    break;
	case 'z':
	    copyzerolength = 0;
	    break;
	case 'n':
	    nfm = atoi(optarg);
	    if (nfm <= 0)
		Error(0, 1, "%s: `-n nfm' must be positive",  prog);
	    break;
	case '?':
	    usage();
	    (void) exit(1);
	}
    }
    if (readtape && writetape)
	Error(0, 1, "%s: Only one of -r and -w allowed.\n", prog);
    else if (!readtape && !writetape)
	Error(0, 1, "%s: One of -r and -w required.\n", prog);
    else if (writetape && nfm > 0)
	Error(0, 1, "%s: -n not allowed with -w.\n", prog);
    else if (writetape && bufsize > 0)
	Error(0, 1, "%s: -b not allowed with -w.\n", prog);
    else if (writetape && copyzerolength == 0)
	Error(0, 1, "%s: -z not allowed with -w.\n", prog);
    if (bufsize < 0)
	bufsize = DEFAULT_BUFSIZE;

    if (readtape)
	readin(bufsize, nfm, copyzerolength, verbose);
    else
	writeout(nfm, verbose);

    return 0;
}

void
readin(bufsize, nfm, copyzerolength, verbose)
int bufsize;
int nfm;
int copyzerolength;
int verbose;
{
    char *buf;
    char smbuf[8];
    long n;
    int i, ifile, nrec;
    long BytesThisFile;

    if (!(buf = (char *) malloc((unsigned) bufsize)))
	Error(1, 1, "%s -r: malloc of %d-long buffer failed:", prog, bufsize);

    if (verbose > 1)
	if (nfm < 0)
	    (void) fprintf(stderr,
	"\n%s -r: number filemarks to copy thru: Double-EOF\nBufsize = %d\n",
		prog, bufsize);
	else
	    (void) fprintf(stderr,
	"\n%s -r: number filemarks to copy thru: %d\nBufsize = %d\n",
		prog, nfm, bufsize);

    /* Send the leader + version */
    i = strlen(PROTOCOL_LEADER);
    if (fwrite(PROTOCOL_LEADER, 1, i, stdout) != i)
	Error(1, 1, "%s -r: write leader failed:", prog);

    i = strlen(PROTOCOL_VERS);
    if (fwrite(PROTOCOL_VERS, 1, i, stdout) != i)
	Error(1, 1, "%s -r: write PROTOCOL_VERS failed:", prog);

    /* Send the buffer size */
    (void) sprintf(smbuf, "%d", bufsize);
    if (fwrite(smbuf, sizeof(smbuf), 1, stdout) != 1)
	Error(1, 1, "%s -r: write bufsize failed:", prog);

    /* Copy to double EOF, if nfm==-1; copy nfm files, otherwise. */
    for (ifile=0, nrec=1; ifile < nfm || (nfm == -1 && nrec > 0); ifile++) {
	/* If this isn't the first file, put an EOF */
	if (ifile > 0) {
	    if (nrec > 0 || copyzerolength) {
		if (verbose) (void) fprintf(stderr,
			"%s -r: EOF %d (%d rec%s; %d bytes)\n",
			prog, ifile, nrec, nrec!=1 ? "s":"", BytesThisFile);
		(void) sprintf(smbuf, "%d", 0);
		if (fwrite(smbuf, sizeof(smbuf), 1, stdout) != 1)
		    Error(1, 1, "%s -r: write EOF indicator failed:", prog);
	    }
	}
	BytesThisFile = 0;
	for (BytesThisFile = 0, nrec=0;
			(n = read(0, buf, bufsize)) > 0;
					    nrec++, BytesThisFile += n) {
	    if (verbose > 1) (void) fprintf(stderr, "%s -r: l=%d\n", prog, n);
	    (void) sprintf(smbuf, "%d", n);
	    if (fwrite(smbuf, sizeof(smbuf), 1, stdout) != 1)
		Error(1, 1, "%s -r: write reclen failed:", prog);
	    if (fwrite(buf, 1, (int) n, stdout) != n)
		Error(1, 1, "%s -r: write of data failed: ", prog);
	}
	if (n < 0) {
	    if (nrec == 0)
		Error(0, 1, "%s -r: Probably reached EOT", prog);
	    else
		Error(1, 1,
			"%s -r: error reading record %d of file %d: ",
			prog, nrec+1, ifile+1);
	}
    }
    if (verbose) (void) fprintf(stderr,
			"%s -r: EOF %d (%d rec%s; %d bytes)\n",
			prog, ifile, nrec, nrec!=1 ? "s":"", BytesThisFile);
}

void
writeout(nfm, verbose)
int nfm;
int verbose;
{
    char *buf;
    char smbuf[8];
    int i, bufsize;
    long n;
    void write_eof();
    int nrec, neof=0;
    char sbuf[100];
    long BytesThisFile;

    /* Check the leader + version */
    i = strlen(PROTOCOL_LEADER);
    if (fread(sbuf, 1, i, stdin) != i)
	Error(1, 1, "%s -w: failed to read leader:", prog);

    if (strncmp(sbuf, PROTOCOL_LEADER, i) != 0)
	Error(0, 1, "%s -w: input isn't in %s form: leader isn't `%s'",
			prog, prog, PROTOCOL_LEADER);

    i = strlen(PROTOCOL_VERS);
    if (fread(sbuf, 1, i, stdin) != i)
	Error(1, 1, "%s -w: failed to read version:", prog);

    if (strncmp(sbuf, PROTOCOL_VERS, i) != 0)
	Error(0, 1, "%s -w: version isn't `%s'", prog, PROTOCOL_VERS);

    /* Get the buffer size */
    if (fread(smbuf, sizeof(smbuf), 1, stdin) != 1)
	Error(1, 1, "%s -w: read bufsize failed: ", prog);

    (void) sscanf(smbuf, "%d", &bufsize);
    if (verbose > 1) (void) fprintf(stderr, "\twrite: bufsize = %d\n", bufsize);
    if (!(buf = (char *) malloc((unsigned) bufsize)))
	Error(1, 1, "%s -w: malloc of %d-long buffer failed", prog, bufsize);

    for (BytesThisFile = 0, nrec=0;
			fread(smbuf, sizeof(smbuf), 1, stdin) == 1; nrec++) {
	(void) sscanf(smbuf, "%d", &n);
	BytesThisFile += n;
	if (n == 0) {
	    neof++;
	    if (verbose) (void) fprintf(stderr,
			"%s -w: EOF %d (%d rec%s; %d bytes)\n",
			prog, neof, nrec, nrec!=1 ? "s":"", BytesThisFile);
	    write_eof(1);	/* Put EOF */
	    nrec = -1;
	    BytesThisFile = 0;
	} else {
	    if (verbose > 1) (void) fprintf(stderr, "\t%s -w: l=%d\n", prog, n);
	    if (fread(buf, 1, (int) n, stdin) != n)
		Error(1, 1, "%s -w: read of data failed: ", prog);

	    if (write(1, buf, (int) n) != n)
		Error(1, 1, "%s -w: write of data failed: ", prog);
	}
    }
    if (ferror(stdin)) {
	Error(1, 1, "%s -w: read of reclength failed: ", prog);
    } else {
	neof++;
	if (verbose) (void) fprintf(stderr,
			"%s -w: EOF %d (%d rec%s; %d bytes)\n",
			prog, neof, nrec, nrec!=1 ? "s":"", BytesThisFile);
    }
}

#ifdef _AIX
void
write_eof(tape_fd)
int tape_fd;
{
    struct stop com;

    com.st_count = 1;
    com.st_op = STWEOF;
    if (ioctl(tape_fd, STIOCTOP, &com) == -1)
	Error(1, 1, "%s -w: Attempt to write EOF failed: ", prog);
}
#else
void
write_eof(tape_fd)
int tape_fd;
{
    struct mtop com;

    com.mt_count = 1;
    com.mt_op = MTWEOF;
    if (ioctl(tape_fd, MTIOCTOP, &com) == -1)
	Error(1, 1, "%s -w: Attempt to write EOF failed: ", prog);
}
#endif
